from acai_aws.apigateway.resolver.cache import ResolverCache
from acai_aws.apigateway.endpoint import Endpoint
from acai_aws.apigateway.exception import ApiException
from acai_aws.apigateway.resolver.modes.directory import DirectoryModeResolver
from acai_aws.apigateway.resolver.modes.mapping import MappingModeResolver
from acai_aws.apigateway.resolver.modes.pattern import PatternModeResolver


class Resolver:
    PATTERN_MODE = 'pattern'
    DIRECTORY_MODE = 'directory'
    MAPPING_MODE = 'mapping'
    __cache_misses = 0
    __available_resolvers = {
        'directory': DirectoryModeResolver,
        'mapping': MappingModeResolver,
        'pattern': PatternModeResolver
    }

    def __init__(self, **kwargs):
        self.__mode = self.__determine_routing_mode(kwargs)
        self.__cacher = ResolverCache(**kwargs)
        self.__resolver = self.__available_resolvers[self.__mode](**kwargs)

    @property
    def cache_misses(self):
        return self.__cache_misses

    def auto_load(self):
        if self.__mode != self.MAPPING_MODE:
            self.__resolver.load_importer_files()

    def get_endpoint(self, request):
        endpoint_module = self.__get_endpoint_module(request)
        if not hasattr(endpoint_module, request.method):
            raise ApiException(code=403, message='method not allowed')
        endpoint = Endpoint(endpoint_module, request.method)
        self.__assign_normalized_route(request, endpoint)
        self.__check_dynamic_route_and_apply_params(request, endpoint)
        self.__cacher.put(request.path, endpoint_module,self.__resolver.has_dynamic_route, self.__resolver.dynamic_parts)
        self.__resolver.reset()
        return endpoint

    def __determine_routing_mode(self, kwargs):
        if isinstance(kwargs['handlers'], dict):
            return self.MAPPING_MODE
        if isinstance(kwargs['handlers'], str) and '*' in kwargs['handlers'] and '.py' in kwargs['handlers']:
            return self.PATTERN_MODE
        return self.DIRECTORY_MODE

    def __get_endpoint_module(self, request):
        cached = self.__cacher.get(request.path)
        endpoint_module = cached.get('endpoint')
        self.__resolver.has_dynamic_route = cached.get('is_dynamic_route', self.__resolver.has_dynamic_route)
        self.__resolver.dynamic_parts = cached.get('dynamic_parts', self.__resolver.dynamic_parts)
        if endpoint_module is None:
            self.__cache_misses += 1
            endpoint_module = self.__resolver.get_endpoint_module(request)
        return endpoint_module

    def __assign_normalized_route(self, request, endpoint):
        base_path_parts = self.__resolver.base_path.split('/')
        dirty_route_parts = endpoint.required_route.split('/') if endpoint.has_required_route else request.path.split('/')
        route_parts = [part for part in dirty_route_parts if part]
        combined_route = base_path_parts + route_parts
        final_route = list(dict.fromkeys(combined_route))
        request.route = '/'.join(final_route)

    def __check_dynamic_route_and_apply_params(self, request, endpoint):
        if not self.__resolver.has_dynamic_route:
            return
        if self.__resolver.has_dynamic_route and not endpoint.has_required_route:
            raise ApiException(
                code=404, 
                key_path=request.path, 
                message='no route found; endpoint does have required_route configured'
            )
        clean_request_path = [rp for rp in request.path.split('/') if rp and rp not in self.__resolver.base_path.split('/')]
        clean_endpoint_route = [er for er in endpoint.required_route.split('/') if er and er not in self.__resolver.base_path.split('/')]
        self.__check_dynamic_route(request, clean_request_path, clean_endpoint_route)
        self.__apply_dynamic_route_params(request, clean_endpoint_route)

    def __check_dynamic_route(self, request, clean_request_path, clean_endpoint_route):
        for index, _ in enumerate(clean_request_path):
            if clean_request_path[index] != clean_endpoint_route[index] and index not in list(self.__resolver.dynamic_parts.keys()):
                raise ApiException(
                    code=404, 
                    key_path=request.path, 
                    message='no route found; requested dynamic route does not match endpoint route definition'
                )

    def __apply_dynamic_route_params(self, request, required_route_parts):
        for part in list(self.__resolver.dynamic_parts.keys()):
            variable_name = required_route_parts[part]
            if not variable_name.startswith('{') or not variable_name.endswith('}'):
                raise ApiException(
                    code=404, 
                    key_path=request.path,
                    message='no route found; endpoint does not have proper variables in required_route'
                )
            dynamic_name = variable_name.strip('{').strip('}')
            request.path_params = dynamic_name, self.__resolver.dynamic_parts[part]



class ApiException(Exception):

    def __init__(self, **kwargs):
        self.code = kwargs.get('code', 500)
        self.key_path = kwargs.get('key_path', 'unknown')
        self.message = kwargs.get('message', 'internal server error')
        super().__init__(self.message)


class ApiTimeOutException(Exception):

    def __init__(self, **kwargs):
        self.code = kwargs.get('code', 408)
        self.key_path = kwargs.get('key_path', 'unknown')
        self.message = kwargs.get('message', 'request timeout')
        super().__init__(self.message)



import unittest

from acai_aws.dynamodb.event import Event
from acai_aws.dynamodb.record import Record
from acai_aws.common.records.exception import RecordException

from tests.mocks.dynamodb import mock_event
from tests.mocks.dynamodb.mock_data_class import MockDDBDataClass


class DynamoDBEventTest(unittest.TestCase):
    created_event = mock_event.get_created_event()
    updated_event = mock_event.get_updated_event()
    deleted_event = mock_event.get_deleted_event()
    schema_path = 'tests/mocks/dynamodb/openapi.yml'
    expected_body = {
        'example_id': '123456789',
        'note': 'Hosrawguw verrig zogupap ce so fajdis vub mos sif mawpowpug kif kihane.',
        'active': True,
        'personal': {
            'gender': 'male',
            'last_name': 'Mcneil',
            'first_name': 'Mannix'
        },
        'transportation': ['public-transit', 'car-access']
    }

    def test_event_accepts_event(self):
        event = Event(self.created_event)
        self.assertEqual(event.context, None)
        self.assertEqual(event.data_class, None)
        self.assertDictEqual(event.event, self.created_event)
        self.assertEqual(len(event.records), len(self.created_event['Records']))

    def test_event_raw_records(self):
        event = Event(self.created_event)
        self.assertCountEqual(event.raw_records, self.created_event['Records'])

    def test_event_returns_record_event(self):
        event = Event(self.created_event)
        self.assertTrue(isinstance(event.records[0], Record))

    def test_event_returns_data_class(self):
        event = Event(self.created_event)
        event.data_class = MockDDBDataClass
        self.assertTrue(isinstance(event.records[0], MockDDBDataClass))
        self.assertTrue(isinstance(event.records[0].record, Record))

    def test_event_validate_record_body_with_schema_file(self):
        event = Event(self.created_event, schema=self.schema_path, required_body='v1-ddb-body')
        self.assertDictEqual(event.records[0].body, self.expected_body)

    def test_event_validate_filters_out_record_body_with_schema_file(self):
        event = Event(self.created_event, schema=self.schema_path, required_body='v1-ddb-body-wrong')
        self.assertEqual(len(event.records), 0)

    def test_event_validate_raises_exception_record_body_with_schema_file(self):
        try:
            event = Event(self.created_event, schema=self.schema_path, required_body='v1-ddb-body-wrong', raise_body_error=True)
            print(event.records)
            self.assertTrue(False)
        except RecordException as record_error:
            self.assertTrue(isinstance(record_error, RecordException))

    def test_event_validate_record_body_with_schema_dict(self):
        schema = {
            '$id': 'https://example.com/person.schema.json',
            '$schema': 'https://json-schema.org/draft/2020-12/schema',
            'type': 'object',
            'required': [
                'example_id',
                'note',
                'active',
                'personal',
                'transportation'
            ],
            'properties': {
                'example_id': {
                    'type': 'string'
                },
                'note': {
                    'type': 'string'
                },
                'active': {
                    'type': 'boolean'
                },
                'personal': {
                    'type': 'object',
                    'properties': {
                        'gender': {
                            'type': 'string'
                        },
                        'first_name': {
                            'type': 'string'
                        },
                        'last_name': {
                            'type': 'string'
                        }
                    }
                },
                'transportation': {
                    'type': 'array',
                    'items': {
                        'type': 'string'
                    }
                }
            }
        }
        event = Event(self.created_event, schema=self.schema_path, required_body=schema)
        self.assertDictEqual(event.records[0].body, self.expected_body)

    def test_event_validate_raises_error_record_body_with_schema_dict(self):
        schema = {
            '$id': 'https://example.com/person.schema.json',
            '$schema': 'https://json-schema.org/draft/2020-12/schema',
            'type': 'object',
            'required': [
                'example_id',
                'note',
                'active',
                'personal',
                'transportation'
            ],
            'properties': {
                'example_id': {
                    'type': 'string'
                },
                'note': {
                    'type': 'boolean'
                },
                'active': {
                    'type': 'boolean'
                },
                'personal': {
                    'type': 'object',
                    'properties': {
                        'gender': {
                            'type': 'string'
                        },
                        'first_name': {
                            'type': 'string'
                        },
                        'last_name': {
                            'type': 'string'
                        }
                    }
                },
                'transportation': {
                    'type': 'array',
                    'items': {
                        'type': 'string'
                    }
                }
            }
        }
        try:
            event = Event(self.created_event, schema=self.schema_path, required_body=schema, raise_body_error=True)
            print(event.records)
            self.assertTrue(False)
        except RecordException as record_error:
            self.assertTrue(isinstance(record_error, RecordException))

    def test_event_validate_filters_out_record_body_with_schema_dict(self):
        schema = {
            '$id': 'https://example.com/person.schema.json',
            '$schema': 'https://json-schema.org/draft/2020-12/schema',
            'type': 'object',
            'required': [
                'example_id',
                'note',
                'active',
                'personal',
                'transportation'
            ],
            'properties': {
                'example_id': {
                    'type': 'string'
                },
                'note': {
                    'type': 'boolean'
                },
                'active': {
                    'type': 'boolean'
                },
                'personal': {
                    'type': 'object',
                    'properties': {
                        'gender': {
                            'type': 'string'
                        },
                        'first_name': {
                            'type': 'string'
                        },
                        'last_name': {
                            'type': 'string'
                        }
                    }
                },
                'transportation': {
                    'type': 'array',
                    'items': {
                        'type': 'string'
                    }
                }
            }
        }
        event = Event(self.created_event, schema=self.schema_path, required_body=schema)
        self.assertEqual(len(event.records), 0)

    def test_event_print(self):
        event = Event(self.created_event)
        try:
            print(event)
            self.assertTrue(True)
        except Exception as error:
            print(error)
            self.assertTrue(False)



from dynamodb_json import json_util as ddb_json

from acai_aws.base.record import BaseRecord


class Record(BaseRecord):

    @property
    def name(self):
        return self._record.get('eventName')

    @property
    def source(self):
        return self._record.get('eventSource')

    @property
    def source_arn(self):
        return self._record.get('eventSourceARN')

    @property
    def version(self):
        return self._record.get('eventVersion')

    @property
    def id(self):
        return self._record.get('eventID')

    @property
    def region(self):
        return self._record.get('awsRegion')

    @property
    def stream_view_type(self):
        return self._record['dynamodb'].get('StreamViewType')

    @property
    def sequence_number(self):
        return self._record['dynamodb'].get('SequenceNumber')

    @property
    def size_bytes(self):
        return self._record['dynamodb'].get('SizeBytes')

    @property
    def keys(self):
        return ddb_json.loads(self._record['dynamodb'].get('Keys', {}))

    @property
    def old_image(self):
        return ddb_json.loads(self._record['dynamodb'].get('OldImage', {}))

    @property
    def new_image(self):
        return ddb_json.loads(self._record['dynamodb'].get('NewImage', {}))

    @property
    def approximate_creation_time(self):
        return self._record['dynamodb'].get('ApproximateCreationDateTime')

    @property
    def body(self):
        return self.new_image

    @property
    def operation(self):
        if self.new_image and not self.old_image:
            return self.CREATED
        if self.new_image and self.old_image:
            return self.UPDATED
        if not self.new_image and self.old_image:
            return self.DELETED
        return self.UNKNOWN

    def __str__(self):
        return str({
            'id': self.id,
            'name': self.name,
            'version': self.version,
            'region': self.region,
            'source_arn': self.source_arn,
            'source': self.source,
            'dynamodb_keys': self.keys,
            'dynamodb_stream_view_type': self.stream_view_type,
            'dynamodb_sequence_number': self.sequence_number,
            'dynamodb_size_bytes': self.size_bytes,
            'dynamodb_old_image': self.old_image,
            'dynamodb_new_image': self.new_image,
            'approximate_creation_time': self.approximate_creation_time
        })



import logging

from acai_aws.common.logger.common_logger import CommonLogger


def log(**kwargs):
    try:
        logger = CommonLogger()
        logger.log(**kwargs)
    except Exception as exception:
        logging.exception(exception)



import os
import unittest

from acai_aws.apigateway.resolver.importer import ResolverImporter
from acai_aws.apigateway.exception import ApiException


class ResolverImporterTest(unittest.TestCase):
    maxDiff = None
    handler_path = 'tests/mocks/apigateway/importer/directory_handlers'
    handler_pattern = 'tests/mocks/apigateway/importer/pattern_handlers/**/*_controller.py'
    handler_bad_multi_dynamic = 'tests/mocks/apigateway/importer/bad_handlers/multi-dynamic'
    handler_bad_same_name = 'tests/mocks/apigateway/importer/bad_handlers/same-name'
    handler_should_pass = 'tests/mocks/apigateway/importer/bad_handlers/should-pass'
    expected_directory_file_tree = {
        '__init__.py': '*',
        '__dynamic_files': {'_dynamic'},
        'basic.py': '*',
        '_dynamic': {
            '__init__.py': '*',
            '__dynamic_files': set()
        },
        'nested_1': {
            '__init__.py': '*',
            '__dynamic_files': set(),
            'nested_2': {
                '__init__.py': '*',
                '__dynamic_files': {'_id.py'},
                '_id.py': '*',
                'basic.py': '*'
            }
        }
    }
    expected_pattern_file_tree = {
        '__dynamic_files': {'_dynamic'},
        'basic_controller.py': '*',
        '_dynamic': {
            '__dynamic_files': set(),
            'dynamic_controller.py': '*'
        },
        'nested_1': {
            '__dynamic_files': set(),
            'nested_2': {
                '__dynamic_files': {'_id_controller.py'},
                'basic_controller.py': '*',
                '_id_controller.py': '*'
            }
        }
    }
    expected_passing_shared_name_diff_levels = {
        '__init__.py': '*',
        '__dynamic_files': set(),
        'good_file.py': '*',
        'good_directory': {
            '__init__.py': '*',
            '__dynamic_files': set(),
            'good_file.py': '*'
        }
    }

    def test_clean_path(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        self.assertEqual(importer.clean_path('/dirty/path/'), 'dirty/path')

    def test_clean_handlers(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        self.assertEqual(importer.handlers, self.handler_path)

    def test_dirty_handlers(self):
        dirty_path = '/tests/mocks/apigateway/importer/directory_handlers/'
        importer = ResolverImporter(handlers=dirty_path, mode='directory')
        self.assertEqual(importer.handlers, self.handler_path)

    def test_handlers_file_tree_directory_mode(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        self.assertDictEqual(self.expected_directory_file_tree, importer.get_handlers_file_tree())

    def test_handlers_file_tree_pattern_mode(self):
        importer = ResolverImporter(handlers=self.handler_pattern, mode='pattern')
        self.assertDictEqual(self.expected_pattern_file_tree, importer.get_handlers_file_tree())

    def test_handlers_file_throw_exception_on_two_dynamic_files(self):
        importer = ResolverImporter(handlers=self.handler_bad_multi_dynamic, mode='directory')
        try:
            print(importer.get_handlers_file_tree())
        except ApiException as importer_error:
            self.assertTrue(isinstance(importer_error, ApiException))
            self.assertTrue('Cannot have two dynamic files in the same directory.' in importer_error.message)

    def test_handlers_file_throw_exception_on_directory_and_file_share_name(self):
        importer = ResolverImporter(handlers=self.handler_bad_same_name, mode='directory')
        try:
            print(importer.get_handlers_file_tree())
            self.assertTrue(False)
        except ApiException as importer_error:
            self.assertTrue(isinstance(importer_error, ApiException))
            self.assertTrue('Cannot have file and directory share same name.' in importer_error.message)

    def test_handlers_file_should_allow_directory_and_file_share_name_on_different_levels(self):
        importer = ResolverImporter(handlers=self.handler_should_pass, mode='directory')
        self.assertDictEqual(self.expected_passing_shared_name_diff_levels, importer.get_handlers_file_tree())

    def test_import_module_from_file(self):
        importer = ResolverImporter(handlers=self.handler_path, mode='directory')
        file_path = f'{self.handler_path}/basic.py'
        import_path = 'tests.mocks.apigateway.importer.directory_handlers.basic'
        handler_module = importer.import_module_from_file(file_path, import_path)
        self.assertTrue(hasattr(handler_module, 'post'))
